#!/bin/bash

# 定义颜色输出
RED='\033[0;31m'
GREEN='\033[0;32m'
YELLOW='\033[1;33m'
BLUE='\033[0;34m'
PURPLE='\033[0;35m'
CYAN='\033[0;36m'
NC='\033[0m' # No Color

# 定义分隔符
SEPARATOR="================================================================"
MINI_SEPARATOR="--------------------------------"

# 检测操作系统
OS_TYPE="unknown"
if [[ "$OSTYPE" == "darwin"* ]]; then
    OS_TYPE="macos"
elif [[ "$OSTYPE" == "linux-gnu"* ]]; then
    OS_TYPE="linux"
fi

# 打印带颜色的消息
print_info() {
    echo -e "${CYAN}[信息]${NC} $1"
}

print_success() {
    echo -e "${GREEN}[成功]${NC} $1"
}

print_error() {
    echo -e "${RED}[错误]${NC} $1"
}

print_warning() {
    echo -e "${YELLOW}[警告]${NC} $1"
}

print_step() {
    echo -e "\n${PURPLE}>>> $1${NC}"
    echo -e "${MINI_SEPARATOR}"
}

# 格式化文件大小（替代 numfmt）
format_size() {
    local size=$1
    local unit=""
    local result=""
    
    if [ $size -ge 1073741824 ]; then
        result=$(echo "scale=2; $size / 1073741824" | bc)
        unit="GiB"
    elif [ $size -ge 1048576 ]; then
        result=$(echo "scale=2; $size / 1048576" | bc)
        unit="MiB"
    elif [ $size -ge 1024 ]; then
        result=$(echo "scale=2; $size / 1024" | bc)
        unit="KiB"
    else
        result=$size
        unit="B"
    fi
    
    echo "${result}${unit}"
}

# 获取文件大小（跨平台）
get_file_size() {
    local file="$1"
    if [ "$OS_TYPE" = "macos" ]; then
        stat -f%z "$file" 2>/dev/null || echo 0
    else
        stat -c%s "$file" 2>/dev/null || echo 0
    fi
}

# 显示帮助信息
show_help() {
    echo -e "${BLUE}${SEPARATOR}${NC}"
    echo -e "${GREEN}RK1106 Flash 烧写工具${NC}"
    echo -e "${BLUE}${SEPARATOR}${NC}"
    echo -e "使用方法: $0 <烧写文件夹路径>"
    echo -e "示例: $0 ~/rv1106/Luckfox_Pico_Max_Flash_250429"
    echo -e "${BLUE}${SEPARATOR}${NC}"
}

# 解析大小字符串（支持 K/M/G 单位）
parse_size() {
    local size_str="$1"
    local size=0
    
    if [[ $size_str =~ ^([0-9]+)([KMG]?)$ ]]; then
        local num="${BASH_REMATCH[1]}"
        local unit="${BASH_REMATCH[2]}"
        
        case "$unit" in
            K) size=$((num * 1024)) ;;
            M) size=$((num * 1024 * 1024)) ;;
            G) size=$((num * 1024 * 1024 * 1024)) ;;
            *) size=$num ;;
        esac
    fi
    
    echo "$size"
}

# 从 env.img 解析分区信息（不使用关联数组）
parse_partitions() {
    local env_file="$1"
    
    # 清空临时文件
    > /tmp/rk_partitions.txt
    > /tmp/rk_addresses.txt
    
    # 读取 env.img 中的 mtdparts 信息
    local mtdparts=$(strings "$env_file" 2>/dev/null | grep -o "mtdparts=[^[:space:]]*" | head -1)
    
    if [ -z "$mtdparts" ]; then
        return 1
    fi
    
    print_info "解析到的分区信息: ${mtdparts#mtdparts=}"
    
    # 提取 spi-nand0 的分区信息
    local nand_parts=$(echo "$mtdparts" | grep -o "spi-nand0:[^[:space:]]*" | cut -d: -f2)
    
    if [ -z "$nand_parts" ]; then
        return 1
    fi
    
    # 解析每个分区
    IFS=',' read -ra PARTS <<< "$nand_parts"
    local current_offset=0
    
    for part in "${PARTS[@]}"; do
        # 解析格式: size@offset(name) 或 size(name)
        if [[ $part =~ ^([0-9]+[KMG]?)@([0-9]+[KMG]?)\(([^\)]+)\)$ ]]; then
            # 格式: size@offset(name)
            local size_str="${BASH_REMATCH[1]}"
            local offset_str="${BASH_REMATCH[2]}"
            local name="${BASH_REMATCH[3]}"
            
            local size=$(parse_size "$size_str")
            local offset=$(parse_size "$offset_str")
            
            echo "$name:$size" >> /tmp/rk_partitions.txt
            echo "$name:$offset" >> /tmp/rk_addresses.txt
            
            # 更新当前偏移量为这个分区的结束位置
            current_offset=$((offset + size))
            
        elif [[ $part =~ ^([0-9]+[KMG]?)\(([^\)]+)\)$ ]]; then
            # 格式: size(name)
            local size_str="${BASH_REMATCH[1]}"
            local name="${BASH_REMATCH[2]}"
            
            local size=$(parse_size "$size_str")
            
            echo "$name:$size" >> /tmp/rk_partitions.txt
            echo "$name:$current_offset" >> /tmp/rk_addresses.txt
            current_offset=$((current_offset + size))
        fi
    done
    
    return 0
}

# 获取分区大小
get_partition_size() {
    local part_name="$1"
    grep "^$part_name:" /tmp/rk_partitions.txt 2>/dev/null | cut -d: -f2
}

# 获取分区地址
get_partition_address() {
    local part_name="$1"
    grep "^$part_name:" /tmp/rk_addresses.txt 2>/dev/null | cut -d: -f2
}

# 检查参数
if [ $# -ne 1 ]; then
    print_error "参数错误！请提供烧写文件夹路径"
    show_help
    exit 1
fi

FLASH_DIR="$1"
SCRIPT_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
RKDEV_TOOL="$SCRIPT_DIR/rkdeveloptool"

echo -e "${BLUE}${SEPARATOR}${NC}"
echo -e "${GREEN}RK1106 Flash 烧写工具 - 开始运行${NC}"
echo -e "${BLUE}${SEPARATOR}${NC}"
print_info "操作系统: $OS_TYPE"

# 检查烧写文件夹是否存在
print_step "检查烧写文件夹"
if [ ! -d "$FLASH_DIR" ]; then
    print_error "烧写文件夹不存在: $FLASH_DIR"
    exit 1
fi
print_success "烧写文件夹路径: $FLASH_DIR"

# 检查 rkdeveloptool 是否存在
print_step "检查烧写工具"
if [ ! -f "$RKDEV_TOOL" ]; then
    print_error "rkdeveloptool 工具不存在！"
    print_info "请先编译 rkdeveloptool 工具并放置在脚本同目录下"
    print_info "工具路径应为: $RKDEV_TOOL"
    exit 1
fi
print_success "找到烧写工具: $RKDEV_TOOL"

# 检查并设置 udev 规则（仅 Linux）
if [ "$OS_TYPE" = "linux" ]; then
    print_step "检查 udev 规则"
    UDEV_RULES="/etc/udev/rules.d/99-rk-rockusb.rules"
    LOCAL_RULES="$SCRIPT_DIR/99-rk-rockusb.rules"
    
    if [ ! -f "$UDEV_RULES" ]; then
        print_warning "系统中未找到 udev 规则文件"
        
        if [ -f "$LOCAL_RULES" ]; then
            print_info "发现本地规则文件: $LOCAL_RULES"
            echo -n -e "${YELLOW}是否将此文件复制到系统 udev 规则目录？[y/N]:${NC} "
            read -r response
            
            if [[ "$response" =~ ^[Yy]$ ]]; then
                print_info "正在复制 udev 规则文件..."
                sudo cp "$LOCAL_RULES" "$UDEV_RULES" || { print_error "复制失败"; exit 1; }
                
                print_info "正在设置文件权限..."
                sudo chmod 644 "$UDEV_RULES" || { print_error "设置权限失败"; exit 1; }
                
                print_info "正在重新加载 udev 规则..."
                sudo udevadm control --reload-rules || { print_error "重载规则失败"; exit 1; }
                sudo udevadm trigger || { print_error "触发规则失败"; exit 1; }
                
                print_success "udev 规则设置完成！"
            else
                print_warning "跳过 udev 规则设置，可能会遇到权限问题"
            fi
        else
            print_error "本地也未找到规则文件: $LOCAL_RULES"
            print_info "请确保 99-rk-rockusb.rules 文件存在于脚本同目录"
        fi
    else
        print_success "udev 规则已存在"
    fi
elif [ "$OS_TYPE" = "macos" ]; then
    print_step "检查系统权限"
    print_info "macOS 不需要 udev 规则"
    print_info "如遇权限问题，请在系统偏好设置->安全性与隐私中允许终端访问"
fi

# 切换到烧写目录
cd "$FLASH_DIR" || exit 1

# 解析 env.img 获取分区信息
print_step "解析分区信息"

if [ -f "env.img" ]; then
    if parse_partitions "env.img"; then
        print_success "成功解析分区信息"
        
        # 显示分区表
        echo -e "\n${CYAN}分区表:${NC}"
        echo -e "${MINI_SEPARATOR}"
        printf "%-15s %-15s %-15s %-15s\n" "分区名" "大小" "起始地址" "扇区地址"
        echo -e "${MINI_SEPARATOR}"
        
        # 创建临时文件用于排序
        > /tmp/rk_partition_table.txt
        
        # 读取所有分区信息
        while IFS=: read -r name addr; do
            size=$(get_partition_size "$name")
            [ -z "$size" ] && continue
            sector=$((addr / 512))
            # 将地址作为第一列，便于排序
            printf "%d\t%-15s\t%-15s\t0x%08X\t0x%X\n" "$addr" "$name" "$(format_size $size)" "$addr" "$sector" >> /tmp/rk_partition_table.txt
        done < /tmp/rk_addresses.txt
        
        # 按第一列（地址）数字排序，然后去掉第一列
        sort -n -k1,1 /tmp/rk_partition_table.txt | while IFS=$'\t' read -r addr name size addr_hex sector_hex; do
            printf "%-15s %-15s %-15s %-15s\n" "$name" "$size" "$addr_hex" "$sector_hex"
        done
        
        echo -e "${MINI_SEPARATOR}"
        
        # 清理临时文件
        rm -f /tmp/rk_partition_table.txt
    else
        print_warning "无法从 env.img 解析分区信息，将使用默认配置"
    fi
fi

# 定义默认烧写地址
declare -a FLASH_ORDER=(
    "env.img:0x0"
    "idblock.img:0x200"
    "uboot.img:0x400"
    "boot.img:0x800"
    "oem.img:0x2800"
    "userdata.img:0x11800"
    "rootfs.img:0x16800"
)

# 如果解析到了分区信息，更新烧写地址
if [ -f /tmp/rk_addresses.txt ]; then
    FLASH_ORDER=("env.img:0x0")  # env.img 总是烧写到 0x0
    
    for img_info in "idblock.img:idblock" "uboot.img:uboot" "boot.img:boot" "oem.img:oem" "userdata.img:userdata" "rootfs.img:rootfs"; do
        img="${img_info%:*}"
        part="${img_info#*:}"
        
        if [ -f "$img" ]; then
            addr=$(get_partition_address "$part")
            if [ -n "$addr" ]; then
                sector=$((addr / 512))
                FLASH_ORDER+=("$img:$(printf "0x%X" $sector)")
            fi
        fi
    done
fi

# 检查必要的镜像文件
print_step "检查镜像文件"
MISSING_FILES=()
EXISTING_FILES=()
OVERSIZED_FILES=()

for entry in "${FLASH_ORDER[@]}"; do
    img="${entry%:*}"
    sector="${entry#*:}"
    
    if [ ! -f "$img" ]; then
        MISSING_FILES+=("$img")
        print_error "缺少文件: $img"
    else
        file_size=$(get_file_size "$img")
        
        # 检查文件大小是否超出分区大小
        case "$img" in
            "env.img") part_name="env" ;;
            "idblock.img") part_name="idblock" ;;
            "uboot.img") part_name="uboot" ;;
            "boot.img") part_name="boot" ;;
            "oem.img") part_name="oem" ;;
            "userdata.img") part_name="userdata" ;;
            "rootfs.img") part_name="rootfs" ;;
            *) part_name="" ;;
        esac
        
        part_size=$(get_partition_size "$part_name")
        if [ -n "$part_size" ] && [ "$part_size" -gt 0 ]; then
            if [ $file_size -gt $part_size ]; then
                OVERSIZED_FILES+=("$img")
                print_error "$img 文件过大: $(format_size $file_size) > 分区大小 $(format_size $part_size)"
            else
                EXISTING_FILES+=("$img:$sector")
                print_success "找到文件: $img -> 扇区 $sector (大小: $(format_size $file_size) / $(format_size $part_size))"
            fi
        else
            EXISTING_FILES+=("$img:$sector")
            print_success "找到文件: $img -> 扇区 $sector (大小: $(format_size $file_size))"
        fi
    fi
done

# 检查 download.bin
if [ ! -f "download.bin" ]; then
    print_error "缺少文件: download.bin (USB插件)"
    MISSING_FILES+=("download.bin")
else
    file_size=$(get_file_size "download.bin")
    print_success "找到文件: download.bin (USB插件) (大小: $(format_size $file_size))"
fi

# 显示文件检查总结
if [ ${#OVERSIZED_FILES[@]} -ne 0 ]; then
    echo -e "\n${RED}以下文件超出分区大小限制:${NC}"
    for img in "${OVERSIZED_FILES[@]}"; do
        echo -e "  ${RED}•${NC} $img"
    done
fi

if [ ${#MISSING_FILES[@]} -ne 0 ] || [ ${#OVERSIZED_FILES[@]} -ne 0 ]; then
    if [ ${#OVERSIZED_FILES[@]} -ne 0 ]; then
        echo -n -e "${RED}有文件超出分区大小限制！${NC}"
    fi
    if [ ${#MISSING_FILES[@]} -ne 0 ]; then
        echo -n -e "${YELLOW}有文件缺失！${NC}"
    fi
    exit 1
fi

# 开始烧写流程
echo -e "\n${BLUE}${SEPARATOR}${NC}"
echo -e "${GREEN}准备开始烧写，请勿断开设备连接！${NC}"
echo -e "${BLUE}${SEPARATOR}${NC}"

# 检查设备连接
print_step "检查设备连接"
print_info "正在搜索 RK 设备..."
DEVICE_INFO=$("$RKDEV_TOOL" ld 2>&1)

if [[ $DEVICE_INFO == *"DevNo"* ]]; then
    print_success "找到设备:"
    echo "$DEVICE_INFO" | grep "DevNo" | while read -r line; do
        echo -e "  ${GREEN}•${NC} $line"
    done
else
    print_error "未检测到设备！请检查:"
    print_info "1. 设备是否正确连接"
    print_info "2. 设备是否处于 Maskrom 模式，需要按住 BOOT 键并上电"
    if [ "$OS_TYPE" = "linux" ]; then
        print_info "3. USB 权限是否正确设置"
    elif [ "$OS_TYPE" = "macos" ]; then
        print_info "3. 系统偏好设置->安全性与隐私中是否允许终端访问"
    fi
    print_info "4. 如果是虚拟机，是否正确配置 USB 设备"
    exit 1
fi

# 烧写 usbplug
print_step "烧写 USB 插件 (download.bin)"
"$RKDEV_TOOL" db download.bin || { print_error "烧写 usbplug 失败，请重新进入 Maskrom 模式后再次尝试"; exit 1; }
print_success "USB 插件烧写完成"
sleep 1

# 获取 Flash 信息
print_step "获取 Flash 信息"
FLASH_INFO=$("$RKDEV_TOOL" rfi 2>&1)
echo "$FLASH_INFO" | grep -E "(Flash Size|Block Size|Page Size|Flash CS)"
print_success "Flash 信息获取完成"

# 执行烧写
echo -e "\n${YELLOW}开始烧写镜像文件...${NC}"
echo -e "${SEPARATOR}"

FAILED_COUNT=0
SUCCESS_COUNT=0

for entry in "${EXISTING_FILES[@]}"; do
    img="${entry%:*}"
    sector="${entry#*:}"
    print_step "烧写 $img -> 扇区 $sector"
    
    # 显示进度
    echo -n -e "${CYAN}正在烧写...${NC} "
    
    # 执行烧写并捕获输出
    OUTPUT=$("$RKDEV_TOOL" wl "$sector" "$img" 2>&1)
    
    if echo "$OUTPUT" | grep -q "Write LBA"; then
        echo -e "${GREEN}✓${NC}"
        print_success "$img 烧写完成"
        ((SUCCESS_COUNT++))
    else
        echo -e "${RED}✗${NC}"
        print_error "$img 烧写失败"
        echo "$OUTPUT" | grep -v "^$" | while read -r line; do
            echo -e "  ${RED}>${NC} $line"
        done
        ((FAILED_COUNT++))
        
        echo -n -e "${YELLOW}是否继续烧写其他镜像？[y/N]:${NC} "
        read -r response
        if [[ ! "$response" =~ ^[Yy]$ ]]; then
            break
        fi
    fi
done

# 清理临时文件
rm -f /tmp/rk_partitions.txt /tmp/rk_addresses.txt

# 显示烧写统计
echo -e "\n${CYAN}烧写统计:${NC}"
echo -e "${MINI_SEPARATOR}"
echo -e "${GREEN}成功: $SUCCESS_COUNT 个文件${NC}"
if [ $FAILED_COUNT -gt 0 ]; then
    echo -e "${RED}失败: $FAILED_COUNT 个文件${NC}"
fi
echo -e "${MINI_SEPARATOR}"

# 复位设备
print_step "复位设备"
print_info "正在复位设备..."
if "$RKDEV_TOOL" rd 2>&1; then
    print_success "设备复位成功"
else
    print_warning "复位命令执行失败，请手动复位设备"
fi

# 完成
echo -e "\n${BLUE}${SEPARATOR}${NC}"
if [ $FAILED_COUNT -eq 0 ]; then
    echo -e "${GREEN}烧写流程完成！所有文件烧写成功${NC}"
else
    echo -e "${YELLOW}烧写流程完成！部分文件烧写失败${NC}"
fi
echo -e "${BLUE}${SEPARATOR}${NC}"
print_info "如遇到问题，请检查:"
print_info "• 设备连接是否正常"
print_info "• 镜像文件是否完整"
if [ "$OS_TYPE" = "linux" ]; then
    print_info "• USB 权限是否正确"
elif [ "$OS_TYPE" = "macos" ]; then
    print_info "• 系统安全设置是否允许终端访问 USB"
fi
print_info "• 如果是虚拟机，是否正确配置 USB 设备"
echo -e "${BLUE}${SEPARATOR}${NC}"